/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package it.eng.javax.jcr.lock;

import it.eng.javax.jcr.Node;
import it.eng.javax.jcr.RepositoryException;

/**
 * Represents a lock placed on an item.
 *
 */
public interface Lock {

    /**
     * Returns the user ID of the user who owns this lock. This is the value of the
     * <code>jcr:lockOwner</code> property of the lock-holding node. It is also the
     * value returned by <code>Session.getUserID</code> at the time that the lock was
     * placed. The lock owner's identity is only provided for informational purposes.
     * It does not govern who can perform an unlock or make changes to the locked nodes;
     * that depends entirely upon who the token holder is.
     * @return a user ID.
     */
    public String getLockOwner();

    /**
     * Returns <code>true</code> if this is a deep lock; <code>false</code> otherwise.
     *
     * @return a boolean
     */
    public boolean isDeep();

    /**
     * Returns the lock holding node. Note that <code>N.getLock().getNode()</code>
     * (where <code>N</code> is a locked node) will only return <code>N</code>
     * if <code>N</code> is the lock holder. If <code>N</code> is in the subtree
     * of the lock holder, <code>H</code>, then this call will return <code>H</code>.
     *
     * @return an <code>Node</code>.
     */
    public Node getNode();

    /**
     * May return the lock token for this lock.
     * <p/>
     * If this <code>Session</code> holds the lock token for this lock, then this method will
     * return that lock token. If this <code>Session</code> does not hold the applicable lock
     * token then this method will return null.
     *
     * @return a <code>String</code>.
     */
    public String getLockToken();

    /**
     * Returns true if this <code>Lock</code> object represents a lock that is currently in effect.
     * If this lock has been unlocked either explicitly or due to an implementation-specific limitation
     * (like a timeout) then it returns <code>false</code>. Note that this method is intended for
     * those cases where one is holding a <code>Lock</code> Java object and wants to find out
     * whether the lock (the JCR-level entity that is attached to the lockable node) that this
     * object originally represented still exists. For example, a timeout or explicit
     * <code>unlock</code> will remove a lock from a node but the <code>Lock</code>
     * Java object corresponding to that lock may still exist, and in that case its
     * <code>isLive</code> method will return <code>false</code>.
     * @return a <code>boolean</code>.
     * @throws RepositoryException if an error occurs.
     */
    public boolean isLive() throws RepositoryException;

    /**
     * Returns <code>true</code> if this is a session-scoped lock.
     * Returns <code>false</code> if this is an open-scoped lock.
     * @return a <code>boolean</code>.
     */
    public boolean isSessionScoped();

    /**
     * If this lock's time-to-live is governed by a timer, this method resets that timer so that the
     * lock does not timeout and expire. If this lock's time-to-live is not governed by a timer,
     * then this method has no effect.
     * <p/>
     * A LockException is thrown if this <code>Session</code> does not hold the correct lock token for this lock.
     * @throws LockException if this <code>Session</code> does not hold the correct lock token for this lock.
     * @throws RepositoryException if another error occurs.
     */
    public void refresh() throws LockException, RepositoryException;
}
