package it.eng.spago.cms.exec;

import java.util.Collection;
import java.util.Iterator;

import it.eng.spago.cms.exceptions.OperationExecutionException;
import it.eng.spago.cms.exec.operations.Operation;
import it.eng.spago.cms.exec.results.ElementDescriptor;
import it.eng.spago.cms.util.constants.ModuleConstants;
import it.eng.spago.error.EMFInternalError;
import it.eng.spago.base.SourceBean;
import it.eng.spago.security.IEngUserProfile;
import it.eng.spago.tracing.TracerSingleton;

public class OperationExecutorImpl implements OperationExecutor {

	public ElementDescriptor getObject(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {
		control(conn, operation, profile);
		GetObjectImpl get = new GetObjectImpl();
		String [] roles = getArrayRoles(profile);
		SourceBean result = get.execute(operation.getOperationDescriptor(), conn, roles);
		ElementDescriptor desc = new ElementDescriptor(result);
		if(closeConn){
			closeConnection(conn);
		}
		return desc;
	}

	public void setObject(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {
		control(conn, operation, profile);
		SetObjectImpl set = new SetObjectImpl();
		String [] roles = getArrayRoles(profile);
		SourceBean result = set.execute(operation.getOperationDescriptor(), conn, roles);
		if(closeConn){
			closeConnection(conn);
		}
	}

	public void deleteObject(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {		
		control(conn, operation, profile);
		DeleteObjectImpl del = new DeleteObjectImpl();
		String [] roles = getArrayRoles(profile);
		del.execute(operation.getOperationDescriptor(), conn, roles);
		if(closeConn){
			closeConnection(conn);
		}
	}

	public void restoreObject(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {
		control(conn, operation, profile);
		RestoreObjectImpl res = new RestoreObjectImpl();
		String [] roles = getArrayRoles(profile);
		res.execute(operation.getOperationDescriptor(), conn, roles);
		if(closeConn){
			closeConnection(conn);
		}
	}

	public SourceBean search(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {
		control(conn, operation, profile);
		SearchObjectImpl search = new SearchObjectImpl();
		String [] roles = getArrayRoles(profile);
		SourceBean resSB = search.execute(operation.getOperationDescriptor(), conn, roles);
		if(closeConn){
			closeConnection(conn);
		}
		return resSB;
	}

	public void renameObject(CMSConnection conn, Operation operation, IEngUserProfile profile, boolean closeConn) throws OperationExecutionException {
		TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.MAJOR,
        		"RenameObject: functionality not implemented");
	}

	
	
	private String [] getArrayRoles(IEngUserProfile profile) throws OperationExecutionException {
		Collection  rolesCol = null;
		try{
		 	rolesCol = profile.getRoles();
		 } catch (Exception e) {
		 	TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                      "OperationExecutorImpl::getArrayRoles: error during profile roles recover operation");
            throw new OperationExecutionException("OperationExecutorImpl::getArrayRoles: error during profile roles recover operation");
		 }
		 String[] roles = new String[rolesCol.size()];
         Iterator iter = rolesCol.iterator();
         int index = 0;
         while(iter.hasNext()) {
         	Object obj = iter.next();
         	roles[index] = obj.toString();
         	index++;
         }
         return roles;         
    }
	
	
	
	/**
	 * Perform controls commons to all the operation
	 */
	private void control(CMSConnection conn, Operation oper, IEngUserProfile profile) throws OperationExecutionException {
		if(profile == null) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: profile null");
            throw new OperationExecutionException("OperationExecutorImpl::control: profile null");
		}
		int numRoles = 0;
		try {
			numRoles = profile.getRoles().size();
		} catch (EMFInternalError eie) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: error during profile roles recover operation");
            throw new OperationExecutionException("OperationExecutorImpl::control: error during profile roles recover operation");
		}
		if(numRoles == 0) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: profile contains no roles");
            throw new OperationExecutionException("OperationExecutorImpl::control: profile contains no roles");
		}
		if(conn == null) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: connection null");
            throw new OperationExecutionException("OperationExecutorImpl::control: connection null");
		}
		if(conn.getSession() == null) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: jcr connection (session) null");
            throw new OperationExecutionException("OperationExecutorImpl::control: jcr connection (session) null ");
		}
		if(conn.isClose()) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: connection close");
            throw new OperationExecutionException("OperationExecutorImpl::control: connection close");
		}
		String prefSys = conn.getPrefixPropertySystem();
		if((prefSys == null) || (prefSys.trim().equals("")) ) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: system prefix not set");
            throw new OperationExecutionException("OperationExecutorImpl::control: system prefix not set");
		}
		String prefUsr = conn.getPrefixPropertySystem();
		if((prefUsr == null) || (prefUsr.trim().equals("")) ) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: user prefix not set");
            throw new OperationExecutionException("OperationExecutorImpl::control: user prefix not set");
		}
		if(oper == null){			
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: operation null");
            throw new OperationExecutionException("OperationExecutorImpl::control: operation null");
	    }
		if(!oper.isComplete()) {
			TracerSingleton.log(ModuleConstants.NAME_MODULE, TracerSingleton.CRITICAL,
                                "OperationExecutorImpl::control: operation descriptor not complete");
            throw new OperationExecutionException("OperationExecutorImpl::control: operation descriptor not complete");
		}
	}
	
	
		
	/**
	 * Close the connection passed
	 * @param conn, connection to close
	 */
	private void closeConnection(CMSConnection conn) {
		if( conn != null && !conn.isClose() ) {
			conn.close();
		}
	}
	
	
	
}
