package it.eng.spago.cms.util;

import it.eng.spago.base.SourceBean;
import it.eng.spago.base.SourceBeanAttribute;
import it.eng.spago.base.SourceBeanException;

import java.util.List;
import java.util.Vector;

/**
 * 
 * Extends the SourceBean class of the Spago framework 
 * in order to add the substituition functionality of params with a
 * context specific value.  
 * The added methods allow: 
 * 
 * <ul>
 * <li>  
 * Replace a vector of parameters to correspondent 
 * attributes of the SourceBean that contain a particular 
 * value, like for example '?'. 
 * </li>
 * <li>
 * Search all attributes of the SourceBean that contain 
 * a particular value.
 * </li>
 * </ul>
 * 
 */

public class ParamSourceBean extends SourceBean {

	
	/**
	 * Build the SourceBean from an existing SourceBean 
	 * 
	 * @param sb SourceBean existing, the new ParamSourceBean will
	 * contains the same data. 
	 * @throws SourceBeanException, throw if the creation of the 
	 * ParamSourceBean fails 
	 */
	public ParamSourceBean(SourceBean sb) throws SourceBeanException
	{
	  super(sb);	
	}
	

	/**
	 * Replace the parameters to all attributes of the ParamSourceBean 
	 * whose value is equal to that specifing in input. 
	 * The substitution happens in parsing hierarchical order of 
	 * the sourceBean. The method controls all branchs of the SourceBean
	 * until the leaves, from the first to the last, and when it finds 
	 * an attribute with the value equals to that in input it replaces 
	 * the value with the value of the next parameter. 
	 * 	  
	 * @param value Value to search in the attributes 
	 * for being replaced from the correspondent value 
	 * of the vector parameters
	 * @param params Vector of params value to insert in the SourceBean 
	 * @return SourceBean with the value of parameters replaced 
	 * @throws SourceBeanException, arise if an error occurs during
	 * sostitution of th attributes value.
	 */
	public synchronized SourceBean replaceParamToken(Object value, Vector params) throws SourceBeanException {
		List attributes = getContainedAttributes();
		int posNextParam = 0;
		for(int i=0; i<attributes.size(); i++)
		 {
			SourceBeanAttribute attribute = (SourceBeanAttribute)attributes.get(i);
			Object val = attribute.getValue();
			if(val instanceof SourceBean){
			  	ParamSourceBean underBean = new ParamSourceBean((SourceBean)val);
			  	attribute.setValue((SourceBean)underBean.replaceParamToken(value, params));
			}
		    else {
		    	try{
		    		if(val.equals(value)) {		    			
		    			attribute.setValue(params.get(0));
		    			params.remove(0);
		    		}
		    	}
		    	catch(Exception e) { /* ignore */ }
		    }
		 }
		return (SourceBean)this;
	}
	/*
	 * NB: the method perform a recursion and returns the object of the stack  
	 * in execution for set it in the father object. 
	 * Could be thought to simply set up the value of the sourceBean, once found, 
	 * without to return the object. Such system does not work because 
	 * of the recursion and the clonation of the method "getContainedAttributes",
	 * that they involve the lacked substitution of the values 
	 * to depth different from 1;
	 * 
	 */

	
	
	/**
	 * 
	 * The method searches all the attributes of the 
	 * ParamSourceBean whose value is equal to that one received
	 * in input. All Attributes found will be insert inside a 
	 * vector.
	 * 
	 * @param value Value to search inside SourceBean attributes 
	 * @return Vector, set of SourceBeanAttribute find 
	 * @throws SourceBeanException, throw if an error occurs
	 * during search operation.
	 * 
	 */
	public synchronized Vector getFilteredSourceBeanAttribute(Object value) throws SourceBeanException
	{
		Vector attrs = new Vector();
		List attributes = getContainedAttributes();
		for(int i=0; i<attributes.size(); i++)
		 {
			SourceBeanAttribute attribute = (SourceBeanAttribute)attributes.get(i);
			Object val = attribute.getValue();
			if(val instanceof SourceBean){
			  	ParamSourceBean underBean = new ParamSourceBean((SourceBean)val);
			  	Vector underAttrs = underBean.getFilteredSourceBeanAttribute(value);
			  	for(int j=0; j<underAttrs.size(); j++)
			  	  {
			  		attrs.add(underAttrs.get(j));
			  	  }
			}
		    else {
		    	try{
		    		//String valStr = (String)val;
		    		if(val.equals(value))
		    			attrs.add(attribute);
		    	}
		    	catch(Exception e) { /* ignore */ }
		    }
		 }
		return attrs;
	}
	
	
}
