/*
 * Copyright 2004-2005 The Apache Software Foundation or its licensors,
 *                     as applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.jackrabbit.core.query.lucene;

import EDU.oswego.cs.dl.util.concurrent.FIFOReadWriteLock;
import org.apache.jackrabbit.core.fs.FileSystemException;
import org.apache.jackrabbit.core.query.AbstractQueryHandler;
import org.apache.jackrabbit.core.query.QueryConstants;
import org.apache.jackrabbit.core.query.ExecutableQuery;
import org.apache.jackrabbit.core.query.QueryHandlerContext;
import org.apache.jackrabbit.core.state.NodeState;
import org.apache.jackrabbit.core.SessionImpl;
import org.apache.jackrabbit.core.ItemManager;
import org.apache.jackrabbit.core.QName;
import org.apache.jackrabbit.core.NoPrefixDeclaredException;
import org.apache.log4j.Logger;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.SortField;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.IndexSearcher;

import it.eng.javax.jcr.RepositoryException;
import it.eng.javax.jcr.query.InvalidQueryException;

import java.io.IOException;

/**
 * Implements a {@link org.apache.jackrabbit.core.query.QueryHandler} using
 * Lucene.
 */
public class SearchIndex extends AbstractQueryHandler {

    /** The logger instance for this class */
    private static final Logger log = Logger.getLogger(SearchIndex.class);

    /**
     * The actual index
     */
    private MultiIndex index;

    /**
     * The analyzer we use for indexing.
     */
    private final Analyzer analyzer;

    /**
     * Read-write lock to synchronize access on the index.
     */
    private final FIFOReadWriteLock readWriteLock = new FIFOReadWriteLock();

    /**
     * Default constructor.
     */
    public SearchIndex() {
        this.analyzer = new StandardAnalyzer();
    }

    /**
     * Initializes this <code>QueryHandler</code>.
     * @throws IOException if an error occurs while initializing this handler.
     */
    protected void doInit() throws IOException {
        try {
            QueryHandlerContext context = getContext();
            index = new MultiIndex(context.getFileSystem(), this,
                    context.getItemStateManager(), context.getRootUUID());
        } catch (FileSystemException e) {
            throw new IOException(e.getMessage());
        }
    }

    /**
     * Adds the <code>node</code> to the search index.
     * @param node the node to add.
     * @throws RepositoryException if an error occurs while indexing the node.
     * @throws IOException if an error occurs while adding the node to the index.
     */
    public void addNode(NodeState node) throws RepositoryException, IOException {
        Document doc = createDocument(node, getNamespaceMappings());
        try {
            readWriteLock.writeLock().acquire();
        } catch (InterruptedException e) {
            throw new RepositoryException("Failed to aquire write lock.");
        }

        try {
            index.addDocument(doc);
        } finally {
            readWriteLock.writeLock().release();
        }
    }

    /**
     * Removes the node with <code>uuid</code> from the search index.
     * @param uuid the UUID of the node to remove from the index.
     * @throws IOException if an error occurs while removing the node from
     * the index.
     */
    public void deleteNode(String uuid) throws IOException {
        Term idTerm = new Term(FieldNames.UUID, uuid);
        try {
            readWriteLock.writeLock().acquire();
        } catch (InterruptedException e) {
            throw new IOException("Failed to aquire write lock.");
        }

        try {
            index.removeDocument(idTerm);
        } finally {
            readWriteLock.writeLock().release();
        }

    }

    /**
     * Creates a new query by specifying the query statement itself and the
     * language in which the query is stated.  If the query statement is
     * syntactically invalid, given the language specified, an
     * InvalidQueryException is thrown. <code>language</code> must specify a query language
     * string from among those returned by QueryManager.getSupportedQueryLanguages(); if it is not
     * then an <code>InvalidQueryException</code> is thrown.
     *
     * @param session the session of the current user creating the query object.
     * @param itemMgr the item manager of the current user.
     * @param statement the query statement.
     * @param language the syntax of the query statement.
     * @throws InvalidQueryException if statement is invalid or language is unsupported.
     * @return A <code>Query</code> object.
     */
    public ExecutableQuery createExecutableQuery(SessionImpl session,
                                             ItemManager itemMgr,
                                             String statement,
                                             String language)
            throws InvalidQueryException {
        return new QueryImpl(session, itemMgr, this,
                getContext().getPropertyTypeRegistry(), statement, language);
    }

    /**
     * Closes this <code>QueryHandler</code> and frees resources attached
     * to this handler.
     */
    public void close() {
        log.info("Closing search index.");
        index.close();
        getContext().destroy();
    }

    /**
     * Executes the query on the search index.
     * @param query the lucene query.
     * @param orderProps name of the properties for sort order.
     * @param orderSpecs the order specs for the sort order properties.
     * <code>true</code> indicates ascending order, <code>false</code> indicates
     * descending.
     * @return the lucene Hits object.
     * @throws IOException if an error occurs while searching the index.
     */
    Hits executeQuery(Query query,
                             QName[] orderProps,
                             boolean[] orderSpecs) throws IOException {
        try {
            readWriteLock.readLock().acquire();
        } catch (InterruptedException e) {
            throw new IOException("Unable to obtain read lock on search index.");
        }

        Hits hits = null;
        try {
            SortField[] sortFields = new SortField[orderProps.length];
            for (int i = 0; i < orderProps.length; i++) {
                String prop = null;
                if (QueryConstants.JCR_SCORE.equals(orderProps[i])) {
                    // order on jcr:score does not use the natural order as
                    // implemented in lucene. score ascending in lucene means that
                    // higher scores are first. JCR specs that lower score values
                    // are first.
                    sortFields[i] = new SortField(null, SortField.SCORE, orderSpecs[i]);
                } else {
                    try {
                        prop = orderProps[i].toJCRName(getNamespaceMappings());
                    } catch (NoPrefixDeclaredException e) {
                        // will never happen
                    }
                    sortFields[i] = new SortField(prop, SharedFieldSortComparator.PROPERTIES, !orderSpecs[i]);
                }
            }

            if (sortFields.length > 0) {
                hits = new IndexSearcher(index.getIndexReader()).search(query, new Sort(sortFields));
            } else {
                hits = new IndexSearcher(index.getIndexReader()).search(query);
            }
        } finally {
            readWriteLock.readLock().release();
        }

        return hits;
    }

    /**
     * Returns the analyzer in use for indexing.
     * @return the analyzer in use for indexing.
     */
    Analyzer getAnalyzer() {
        return analyzer;
    }

    /**
     * Returns the namespace mappings for the internal representation.
     * @return the namespace mappings for the internal representation.
     */
    NamespaceMappings getNamespaceMappings() {
        return index.getNamespaceMappings();
    }

    /**
     * Creates a lucene <code>Document</code> from a node state using the
     * namespace mappings <code>nsMappings</code>.
     * @param node the node state to index.
     * @param nsMappings the namespace mappings of the search index.
     * @return a lucene <code>Document</code> that contains all properties
     *  of <code>node</code>.
     * @throws RepositoryException if an error occurs while indexing the
     *  <code>node</code>.
     */
    protected Document createDocument(NodeState node, NamespaceMappings nsMappings)
            throws RepositoryException {
        return NodeIndexer.createDocument(node, getContext().getItemStateManager(),
                nsMappings);
    }

    //--------------------------< properties >----------------------------------

    public void setUseCompoundFile(boolean b) {
        index.setUseCompoundFile(b);
    }

    public void setMinMergeDocs(int minMergeDocs) {
        index.setMinMergeDocs(minMergeDocs);
    }

    public void setMaxMergeDocs(int maxMergeDocs) {
        index.setMaxMergeDocs(maxMergeDocs);
    }

    public void setMergeFactor(int mergeFactor) {
        index.setMergeFactor(mergeFactor);
    }

    /**
     * @deprecated redo size is equivalent to minMergeDocs. Use
     * {@link #setMinMergeDocs(int)} to set the size of the redo log.
     */
    public void setRedoSize(int size) {
    }

    public void setBufferSize(int size) {
        index.setBufferSize(size);
    }
}
